import The from "./The.js";
import RequestList from "./RequestList.js";
import ZRaySocket from "./ZRaySocket.js"
import State from "./State.js";
import Collapsible from "./elements/Collapsible.js";
import Sortable from "./elements/Sortable.js";
import Tabable from "./elements/Tabable.js";
import dom from "./elements/Dom.js";

let version = [1,2];
let tabId = 0;

// console.log('start', typeof chrome, typeof chrome.devtools);
if (typeof chrome !== 'undefined' && typeof chrome.devtools !== 'undefined') {
	version = chrome.app.getDetails().version.split('.');
	tabId = chrome.devtools.inspectedWindow.tabId;
}

const apiVersion = { major: parseInt(version[0]), minor: parseInt(version[1]) };
const container = new The({
	tabId,
	token: '',
	host: '',
	protocol: '',
	timing: {}
});

const template = {
	toolbar: () => `
		<div class="toolbar-container">
			<div class="toolbar">
				<button class="icon-btn listen" data-action="toggle" title="Activate">
					<span id="activate" class="activate-icon"></span>
				</button>
				<button class="icon-btn listen" data-action="clearRequests" title="Clear">
					<span class="clear-icon"></span>
				</button>
				<div class="toolbar-separator"></div>
				<!--
				<input type=text value="filter" disabled=true>
				<button class="icon-btn listen" title="Filter">
					<span class="filter-icon"></span>
				</button>
				-->
			</div>
			<div class="toolbar hidden" id="token-toolbar">
				<div class="toolbar-separator"></div>
				<span>Token:</span>
				<span id="copy-token">***</span>
				<button class="icon-btn listen" data-action="copyToken" title="Copy">
					<span class="copy-icon"></span>
				</button>
			</div>
			<div class="toolbar">
				<div class="toolbar-separator"></div>
				<span title="Version">${version.join('.')}&nbsp;</span>
			</div>
		</div>`,
	about: () => `
		<ul id="list-root">
			<li>Chromium extension for <b>Z-Ray</b> by <a href="https://perforce.com" target="_blank">Perforce</a></li>
			<li>Version: ${version.join('.')}</li>
		</ul>`,
	version: {
		err: (version) => `
			<div class="warning">
				<h1>Version error</h1>
				<p>Z-Ray API version ${version} on the server requires similar version of Z-Ray extension</p>
			</div>`,
		warn: (version) => `
			<div class="warning">
				<h1>Version warning</h1>
				<p>Z-Ray API version ${version} on the server does not match this extension version</p>
			</div>`,
	}
}

class ZRay {

	constructor() {
		// console.log("Z-Ray tabId", The.tabId());

		let body = document.getElementsByTagName('body')[0];
		dom.create(template.toolbar(), { appendTo: body });
		this.state = new State(tabId);
		this.requestList = false;

		this.content = dom.div({ class: "widget hbox", appendTo: body });

		// load About page
		dom.create(template.about(), { appendTo: this.content });

		this.isActive = false;
		this.requestStack = [];

		// listen all clicks
		document.addEventListener(
			'click',
			(event) => {
				// console.log('click', event.target);
				let element = event.target;
				if (element.nodeName == 'CODE') {
					this.codeToggle(element);
					return;
				}
				do {
					if (element.classList && element.matches('.listen')) {
						break;
					}
					if (element.nodeName == 'BODY') return;
					element = element.parentNode || false;
				}
				while (element);

				// If the clicked element doesn't have the right selector, bail
				if (!element) {
					// console.log('not clickable');
					return;
				}
				if (!element.dataset.action) {
					// console.log('missing data-action', event);
					return;
				}
				// Don't follow the link
				event.preventDefault();
				// execute this.<method>
				this[element.dataset.action](element);
			},
			false
		);

		if (tabId) {
			// query current url
			chrome.tabs.query(
				{ active: true, lastFocusedWindow: true },
				(tabs) => {
					let url = tabs[0].url;
					// use `url` here inside the callback because it's asynchronous!
					this.matchHost(url);
				}
			);

			// listen url navigation
			chrome.devtools.network.onNavigated.addListener(
				(url) => {
					this.matchHost(url);
				}
			);
		}
		else {
			this.matchHost(document.location.href);
		}
	}

	// monitor browser location changes and disconnect WS when host changes
	matchHost(url) {
		// console.log({match: url, self:this});
		let match = url.match(/([a-z]+):\/\/(.[^/:]+)/i);
		if (match != null && match.length > 1 && typeof match[2] === 'string' && match[2].length > 0) {
			let host = match[2];
			// console.log('host match', match);

			if (host !== The.host() && this.isActive) {
				// navigated away
				// console.log('host change');
				this.stop();
			}
			container.set('host', host);
			container.set('protocol', match[1]);
		}
	}

	start() {
		if (The.host().length == 0) {
			console.log('No host to connect to');
			return;
		}
		this.isActive = true;
		container.set('token', '');
		this.state.setConnecting();

		console.log('ZRay connect to ', The.host());

		this.zrayd = new ZRaySocket(tabId, "ws://"+The.host()+":10091/");
		this.zrayd.connect(false);

		this.zrayd.onmessage = (event) => { this._event(event.data) };
		this.zrayd.onclose = (event) => { this._close(event.data) };

		if (tabId) {
			// delete ZRayTok cookie for current url
			chrome.cookies.remove({'url': The.protocol() + "://" + The.host(), 'name': "zraytok"}, function() {});

			// maybe this addListener has to be inside ws.onopen
			//TODO: this will change with manifest 3.0
			chrome.webRequest.onBeforeSendHeaders.addListener(
				(request) => {
					// console.log({ name: 'z_ray req mod', value: The.token() });
					if (this.isActive && tabId == request.tabId && The.token().length > 0) {
						request.requestHeaders.push({ name: 'zraytok', value: The.token() }); // value must be a string

						// console.log('request', request.url);
						container.setRequestTiming(request.url);

						return { requestHeaders: request.requestHeaders };
					}
				},
				{ urls: ["*://" + The.host() + "/*"] },
				["blocking", "requestHeaders"]
			);
		}
	}

	// disconnect websocket
	stop() {
		if (this.isActive) {
			this.isActive = false;
			// console.log('stopping');
			this.zrayd.close();
			if (tabId) {
				chrome.webRequest.onBeforeSendHeaders.removeListener(this);
			}
			this.state.setInactive();
		}
	}

	// clicks on activation btn
	toggle() {
		// console.log('toggle');
		// console.log(The.token());
		this.isActive ? this.stop() : this.start();
	}

	// clicks on request tabs
	showTab(btn) {
		//console.log('show tab');
		Tabable.click(btn);
	}

	// clicks in requests list
	showRequest(row) {
		//console.log('show request');
		if (row.classList && row.classList.contains('focus')) {
			return;
		}
		this.requestList.show(row);
	}

	// clicks on clear requests button
	clearRequests() {
		// console.log('clear requests');
		if (this.requestList) this.requestList.clearAll();
	}

	// clicks on token copy button
	copyToken(btn) {
		State.copyToken(btn);
	}

	// toggle clicked collapsible (Tree) element
	collapsible(btn) {
		Collapsible.toggle(btn);
	}

	// toggle clicked sortable (table header) element data order
	sortable(btn) {
		Sortable.sort(btn);
	}

	// toggle linefeeds in <code> blocks
	codeToggle(code) {
		let selectionDetection = window.getSelection();
		if (selectionDetection.type == 'Range') return;
		let text = code.innerText;
		let splitter = code.splitter || false;
		if (!splitter) {
			let maxCount = 2;
			splitter = '!';
			let chars = [',', ';', '-'];
			for(let c in chars) {
				let count = text.split(chars[c]).length - 1;
				if (count > maxCount) {
					maxCount = count;
					splitter = chars[c];
				}
			}
			code.splitter = splitter;
		}
		if (splitter != '!') {
			if (code.classList.contains('code-pre')) {
				code.innerText = text.replaceAll(splitter + "\n", splitter);
			}
			else {
				code.innerText = text.replaceAll(splitter, splitter + "\n");
			}
			code.classList.toggle('code-pre');
		}
	}

	// json decode with error check
	_decode(jsonString) {
		let data = {}
		try {
			data = JSON.parse(jsonString)
		}
		catch (error) {
			console.error(error + '-' + jsonString);
		}

		return data;
	}

	// WS close event handler
	_close(str) {
		// console.log('lost connection', this.isActive);
		if (this.isActive) {
			// unexpected disconnect, try to reconnect
			this.state.setConnecting();
		}
	}

	// event from WS queue
	_event(jsonString) {
		let data = this._decode(jsonString);

		// console.log(data);
		if (!data.type) {
			console.log('ZRay::_event - unknown event', jsonString);
			return
		}
		// console.log(data.type, jsonString);
		// first event after connecting to WS queue
		if (data.type == "HELLO") {
			this._initQueue(data);
		}
		else {
			if (!data.request_id) {
				console.log('ZRay::_event - unidentified request', jsonString.slice(0,32));
				return
			}
			if (!this.requestList) {
				// init request list
				this.requestList = new RequestList();
				// delete "about" content
				this.content.innerHTML = '';
				this.content.appendChild(this.requestList.get());
			}
			// send WS data to RequestList
			this.requestList.add(data);
		}
	}

	_initQueue(data) {
		// console.log("queue_id", data.queue_id);
		// API Version control
		let major = data.major_version || "0";
		let minor = data.minor_version || "0";
		if (isNaN(major) || apiVersion.major !== parseInt(major)) {
			// load error page
			this.content.innerHTML = '';
			dom.create(
				template.version.err(`${major}.${minor}`),
				{ appendTo: this.content }
			);
			console.log("Unsupported version");
			this.stop();
			return;
		}
		if (apiVersion.minor !== parseInt(minor)) {
			// load Warning page
			this.content.innerHTML = '';
			dom.create(
				template.version.warn(`${major}.${minor}`),
				{ appendTo: this.content }
			);
		}

		container.set('token', data.queue_id);
		this.state.setActive(data.queue_id);
	}
}

const zray = new ZRay();